#include <math.h>

#include "YRTTriangle.h"
#include "YRTUtils.h"

#pragma hdrstop

bool TYRTTriangle::GetIntersection(TYRTRay ARay, TYRTVector *APoint, TColor *AColor)
{
    if (IsDegenerate)
        return false;

    // Note: if the following instruction is moved down, we lose about 5% on speed...
    TYRTVector w0 = ARay.Start - Points[0];

    // Check if the ray is parallel to the triangle plane:
    float b = dot(norm, ARay.Dir);
    if (fabs(b) < EPSILON)
        return false;

    // Check if the ray goes away from the triangle:
    float a = -dot(norm, w0);
    if((a < 0 && b > 0) || (a > 0 && b < 0))
        return false;

    // The intersection point of the ray with the triangle plane:
    *APoint = ARay.Start + (a / b * ARay.Dir);

    // Check whether APoint is inside the triangle - get and test parametric coordinates:
    TYRTVector w = *APoint - Points[0];

    float wu = dot(w, u);
    float wv = dot(w, v);
    float s = (uv * wv - vv * wu);
    
    if (D >= 0)
    {
        if (s < 0.0 || s > D)
            return false;
        float t = (uv * wu - uu * wv);
        if (t < 0.0 || (s + t) > D)
            return false;
    }
    else
    {
        if (s > 0.0 || s < D)
            return false;
        float t = (uv * wu - uu * wv);
        if (t > 0.0 || (s + t) < D)
            return false;
    }

    *AColor = this->Color;
    return true;
}

void TYRTTriangle::RunPrecalculations(void)
{
    // Get triangle edge vectors and the plane normal:
    u = Points[1] - Points[0];
    v = Points[2] - Points[0];

    norm = u * v;
    IsDegenerate = norm.IsNearZero();

    uu = dot(u, u);
    uv = dot(u, v);
    vv = dot(v, v);

    D = uv * uv - uu * vv;
}

